/* This file is part of Zanshin

   Copyright 2014 Kevin Ottens <ervin@kde.org>
   Copyright 2014 Franck Arrecot <franck.arrecot@gmail.com>

   This program is free software; you can redistribute it and/or
   modify it under the terms of the GNU General Public License as
   published by the Free Software Foundation; either version 2 of
   the License or (at your option) version 3 or any later version
   accepted by the membership of KDE e.V. (or its successor approved
   by the membership of KDE e.V.), which shall act as a proxy
   defined in Section 14 of version 3 of the license.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301,
   USA.
*/

#include "akonaditagrepository.h"

#include "akonadiitemfetchjobinterface.h"
#include "akonadiserializer.h"
#include "akonadistorage.h"
#include "akonaditagcreatejobinterface.h"

#include "utils/compositejob.h"

using namespace Akonadi;


TagRepository::TagRepository(QObject *parent)
    : QObject(parent),
      m_storage(new Storage),
      m_serializer(new Serializer),
      m_ownInterfaces(true)
{
}

TagRepository::TagRepository(StorageInterface *storage, SerializerInterface *serializer)
    : m_storage(storage),
      m_serializer(serializer),
      m_ownInterfaces(false)
{
}

TagRepository::~TagRepository()
{
    if (m_ownInterfaces) {
        delete m_storage;
        delete m_serializer;
    }
}

KJob *TagRepository::create(Domain::Tag::Ptr tag)
{
    auto akonadiTag = m_serializer->createAkonadiTagFromTag(tag);
    Q_ASSERT(!akonadiTag.isValid());
    return m_storage->createTag(akonadiTag)->kjob();
}

KJob *TagRepository::createOrMergeAndAssociate(Domain::Tag::Ptr tag, Domain::Artifact::Ptr child)
{
    auto akonadiTag = m_serializer->createAkonadiTagFromTag(tag);
    Q_ASSERT(!akonadiTag.isValid());
    auto job = new Utils::CompositeJob();
    Akonadi::TagCreateJobInterface *tagCreateJob = m_storage->createTag(akonadiTag, true);
    job->install(tagCreateJob->kjob(), [akonadiTag, tagCreateJob, job, child, this] {
        if (tagCreateJob->kjob()->error() != KJob::NoError)
            return;

        auto associateJob = associate(m_serializer->createTagFromAkonadiTag(tagCreateJob->tag()), child);
        job->addSubjob(associateJob);
    });

    return job;
}

KJob *TagRepository::remove(Domain::Tag::Ptr tag)
{
   auto akonadiTag = m_serializer->createAkonadiTagFromTag(tag);
   Q_ASSERT(akonadiTag.isValid());
   return m_storage->removeTag(akonadiTag);
}

KJob *TagRepository::associate(Domain::Tag::Ptr parent, Domain::Artifact::Ptr child)
{
    auto akonadiTag = m_serializer->createAkonadiTagFromTag(parent);
    Q_ASSERT(akonadiTag.isValid());

    Item childItem;
    if (auto task = child.objectCast<Domain::Task>())
        childItem = m_serializer->createItemFromTask(task);
    else if (auto note = child.objectCast<Domain::Note>())
        childItem = m_serializer->createItemFromNote(note);

    Q_ASSERT(childItem.isValid());

    auto job = new Utils::CompositeJob();
    ItemFetchJobInterface *fetchItemJob = m_storage->fetchItem(childItem);
    job->install(fetchItemJob->kjob(), [akonadiTag, fetchItemJob, parent, job, this] {
        if (fetchItemJob->kjob()->error() != KJob::NoError)
            return;

        Q_ASSERT(fetchItemJob->items().size() == 1);
        auto childItem = fetchItemJob->items().first();
        childItem.setTag(akonadiTag);

        auto updateJob = m_storage->updateItem(childItem);
        job->addSubjob(updateJob);
        updateJob->start();
    });
    return job;
}

KJob *TagRepository::dissociate(Domain::Tag::Ptr parent, Domain::Artifact::Ptr child)
{
    Item childItem;

    if (Domain::Task::Ptr task = child.dynamicCast<Domain::Task>())
        childItem = m_serializer->createItemFromTask(task);
    else if (Domain::Note::Ptr note= child.dynamicCast<Domain::Note>())
        childItem = m_serializer->createItemFromNote(note);

    Q_ASSERT(childItem.isValid());

    auto job = new Utils::CompositeJob();
    ItemFetchJobInterface *fetchItemJob = m_storage->fetchItem(childItem);
    job->install(fetchItemJob->kjob(), [fetchItemJob, parent, job, this] {
        if (fetchItemJob->kjob()->error() != KJob::NoError)
            return;

        Q_ASSERT(fetchItemJob->items().size() == 1);
        auto childItem = fetchItemJob->items().first();
        auto tag = m_serializer->createAkonadiTagFromTag(parent);
        Q_ASSERT(tag.isValid());
        childItem.clearTag(tag);

        auto updateJob = m_storage->updateItem(childItem);
        job->addSubjob(updateJob);
        updateJob->start();
    });

    return job;
}

KJob *TagRepository::dissociate(Domain::Artifact::Ptr child)
{
    Item childItem;

    if (Domain::Task::Ptr task = child.dynamicCast<Domain::Task>())
        childItem = m_serializer->createItemFromTask(task);
    else if (Domain::Note::Ptr note= child.dynamicCast<Domain::Note>())
        childItem = m_serializer->createItemFromNote(note);

    Q_ASSERT(childItem.isValid());

    auto job = new Utils::CompositeJob();
    ItemFetchJobInterface *fetchItemJob = m_storage->fetchItem(childItem);
    job->install(fetchItemJob->kjob(), [fetchItemJob, job, this] {
        if (fetchItemJob->kjob()->error() != KJob::NoError)
            return;

        Q_ASSERT(fetchItemJob->items().size() == 1);
        auto childItem = fetchItemJob->items().first();
        childItem.clearTags();

        auto updateJob = m_storage->updateItem(childItem);
        job->addSubjob(updateJob);
        updateJob->start();
    });

    return job;
}
